@@ -1,12 +1,11 @@
 """CmdStanModel"""
 
+import logging
 import os
 import platform
 import re
-import subprocess
 import shutil
-import logging
-
+import subprocess
 from collections import OrderedDict
 from concurrent.futures import ThreadPoolExecutor
 from multiprocessing import cpu_count
@@ -16,42 +15,60 @@ from typing import Any, Dict, List, Union
 
 from cmdstanpy.cmdstan_args import (
     CmdStanArgs,
-    SamplerArgs,
-    OptimizeArgs,
     GenerateQuantitiesArgs,
+    OptimizeArgs,
+    SamplerArgs,
     VariationalArgs,
 )
 from cmdstanpy.compiler_opts import CompilerOptions
 from cmdstanpy.stanfit import (
-    RunSet,
+    CmdStanGQ,
     CmdStanMCMC,
     CmdStanMLE,
-    CmdStanGQ,
     CmdStanVB,
+    RunSet,
 )
 from cmdstanpy.utils import (
-    do_command,
     EXTENSION,
-    cmdstan_path,
     MaybeDictToFilePath,
     TemporaryCopiedFile,
+    cmdstan_path,
+    do_command,
     get_logger,
     scan_sampler_csv,
 )
 
 
 class CmdStanModel:
+    # overview, omitted from doc comment in order to improve Sphinx docs.
+    #    A CmdStanModel object encapsulates the Stan program and provides
+    #    methods for compilation and inference.
     """
-    Stan model.
+    The constructor method allows model instantiation given either the
+    Stan program source file or the compiled executable, or both.
+    By default, the constructor will compile the Stan program on instantiation
+    unless the argument ``compile=False`` is specified.
+    The set of constructor arguments are:
 
-    - Stores pathnames to Stan program, compiled executable, and collection of
-      compiler options.
-    - Provides functions to compile the model and perform inference on the
-      model given data.
-    - By default, compiles model on instantiation - override with argument
-      ``compile=False``
-    - By default, property ``name`` corresponds to basename of the Stan program
-      or exe file - override with argument ``model_name=<name>``.
+    :param model_name: Model name, used for output file names.
+        Optional, default is the base filename of the Stan program file.
+
+    :param stan_file: Path to Stan program file.
+
+    :param exe_file: Path to compiled executable file.  Optional, unless
+        no Stan program file is specified.  If both the program file and
+        the compiled executable file are specified, the base filenames
+        must match, (but different directory locations are allowed).
+
+    :param compile: Whether or not to compile the model.  Default is ``True``.
+
+    :param stanc_options: Options for stanc compiler, specified as a Python
+        dictionary containing Stanc3 compiler option name, value pairs.
+        Optional.
+
+    :param cpp_options: Options for C++ compiler, specified as a Python
+        dictionary containing C++ compiler option name, value pairs.
+        Optional.
     """
 
     def __init__(
@@ -64,7 +81,17 @@ class CmdStanModel:
         cpp_options: Dict = None,
         logger: logging.Logger = None,
     ) -> None:
-        """Initialize object."""
+        """
+        Initialize object given constructor args.
+
+        :param model_name: Model name, used for output file names.
+        :param stan_file: Path to Stan program file.
+        :param exe_file: Path to compiled executable file.
+        :param compile: Whether or not to compile the model.
+        :param stanc_options: Options for stanc compiler.
+        :param cpp_options: Options for C++ compiler.
+        :param logger: Python logger object.
+        """
         self._name = None
         self._stan_file = None
         self._exe_file = None
@@ -161,7 +188,7 @@ class CmdStanModel:
         repr = 'CmdStanModel: name={}'.format(self._name)
         repr = '{}\n\t stan_file={}'.format(repr, self._stan_file)
         repr = '{}\n\t exe_file={}'.format(repr, self._exe_file)
-        repr = '{}\n\t compiler_optons={}'.format(repr, self._compiler_options)
+        repr = '{}\n\t compiler_options={}'.format(repr, self._compiler_options)
         return repr
 
     @property
@@ -169,7 +196,7 @@ class CmdStanModel:
         """
         Model name used in output filename templates. Default is basename
         of Stan program or exe file, unless specified in call to constructor
-        via argument `model_name`.
+        via argument ``model_name``.
         """
         return self._name
 
@@ -190,7 +217,7 @@ class CmdStanModel:
 
     @property
     def cpp_options(self) -> Dict:
-        """Options to c++ compilers."""
+        """Options to C++ compilers."""
         return self._compiler_options._cpp_options
 
     def code(self) -> str:
@@ -228,7 +255,8 @@ class CmdStanModel:
             ``#include`` directives in order to force recompilation when changes
             are made to the included files.
 
-        :param compiler_options: Options for stanc and C++ compilers.
+        :param stanc_options: Options for stanc compiler.
+        :param cpp_options: Options for C++ compiler.
 
         :param override_options: When ``True``, override existing option.
             When ``False``, add/replace existing options.  Default is ``False``.
@@ -313,6 +341,7 @@ class CmdStanModel:
         seed: int = None,
         inits: Union[Dict, float, str] = None,
         output_dir: str = None,
+        sig_figs: int = None,
         algorithm: str = None,
         init_alpha: float = None,
         iter: int = None,
@@ -366,6 +395,11 @@ class CmdStanModel:
             files are written. If unspecified, output files will be written
             to a temporary directory which is deleted upon session exit.
 
+        :param sig_figs: Numerical precision used for output CSV and text files.
+            Must be an integer between 1 and 18.  If unspecified, the default
+            precision for the system file I/O is used; the usual value is 6.
+            Introduced in CmdStan-2.25.
+
         :param algorithm: Algorithm to use. One of: 'BFGS', 'LBFGS', 'Newton'
 
         :param init_alpha: Line search step size for first iteration
@@ -387,6 +421,7 @@ class CmdStanModel:
                 seed=seed,
                 inits=_inits,
                 output_dir=output_dir,
+                sig_figs=sig_figs,
                 save_diagnostics=False,
                 method_args=optimize_args,
             )
@@ -401,6 +436,7 @@ class CmdStanModel:
         mle = CmdStanMLE(runset)
         return mle
 
+    # pylint: disable=too-many-arguments
     def sample(
         self,
         data: Union[Dict, str] = None,
@@ -424,6 +460,7 @@ class CmdStanModel:
         adapt_step_size: int = None,
         fixed_param: bool = False,
         output_dir: str = None,
+        sig_figs: int = None,
         save_diagnostics: bool = False,
         show_progress: Union[bool, str] = False,
         validate_csv: bool = True,
@@ -501,7 +538,8 @@ class CmdStanModel:
         :param save_warmup: When ``True``, sampler saves warmup draws as part of
             the Stan csv output file.
 
-        :param thin: Period between saved samples.
+        :param thin: Period between recorded iterations.  Default is 1, i.e.,
+             all iterations are recorded.
 
         :param max_treedepth: Maximum depth of trees evaluated by NUTS sampler
             per iteration.
@@ -521,13 +559,13 @@ class CmdStanModel:
             length must match the number of chains and all paths must be
             unique.
 
-        :param step_size: Initial stepsize for HMC sampler.  The value is either
-            a single number or a list of numbers which will be used as the
-            global or per-chain initial step size, respectively.
+        :param step_size: Initial step size for HMC sampler.  The value is
+            either a single number or a list of numbers which will be used
+            as the global or per-chain initial step size, respectively.
             The length of the list of step sizes must match the number of
             chains.
 
-        :param adapt_engaged: When True, adapt stepsize and metric.
+        :param adapt_engaged: When True, adapt step size and metric.
 
         :param adapt_delta: Adaptation target Metropolis acceptance rate.
             The default value is 0.8.  Increasing this value, which must be
@@ -540,7 +578,7 @@ class CmdStanModel:
             towards the typical set.
 
         :param adapt_metric_window: The second phase of adaptation tunes
-            the metric and stepsize in a series of intervals.  This parameter
+            the metric and step size in a series of intervals.  This parameter
             specifies the number of iterations used for the first tuning
             interval; window size increases for each subsequent interval.
 
@@ -562,6 +600,11 @@ class CmdStanModel:
             files are written. If unspecified, output files will be written
             to a temporary directory which is deleted upon session exit.
 
+        :param sig_figs: Numerical precision used for output CSV and text files.
+            Must be an integer between 1 and 18.  If unspecified, the default
+            precision for the system file I/O is used; the usual value is 6.
+            Introduced in CmdStan-2.25.
+
         :param save_diagnostics: Whether or not to save diagnostics. If True,
             csv output files are written to an output file with filename
             template '<model_name>-<YYYYMMDDHHMM>-diagnostic-<chain_id>',
@@ -682,6 +725,7 @@ class CmdStanModel:
                 seed=seed,
                 inits=_inits,
                 output_dir=output_dir,
+                sig_figs=sig_figs,
                 save_diagnostics=save_diagnostics,
                 method_args=sampler_args,
                 refresh=refresh,
@@ -753,6 +797,7 @@ class CmdStanModel:
         mcmc_sample: Union[CmdStanMCMC, List[str]] = None,
         seed: int = None,
         gq_output_dir: str = None,
+        sig_figs: int = None,
     ) -> CmdStanGQ:
         """
         Run CmdStan's generate_quantities method which runs the generated
@@ -796,6 +841,11 @@ class CmdStanModel:
             files are saved.  If unspecified, files will be written to a
             temporary directory which is deleted upon session exit.
 
+        :param sig_figs: Numerical precision used for output CSV and text files.
+            Must be an integer between 1 and 18.  If unspecified, the default
+            precision for the system file I/O is used; the usual value is 6.
+            Introduced in CmdStan-2.25.
+
         :return: CmdStanGQ object
         """
         sample_csv_files = []
@@ -804,7 +854,7 @@ class CmdStanModel:
 
         if isinstance(mcmc_sample, CmdStanMCMC):
             sample_csv_files = mcmc_sample.runset.csv_files
-            sample_drawset = mcmc_sample.draws_as_dataframe()
+            sample_drawset = mcmc_sample.draws_pd()
             chains = mcmc_sample.chains
             chain_ids = mcmc_sample.chain_ids
         elif isinstance(mcmc_sample, list):
@@ -849,7 +899,7 @@ class CmdStanModel:
                 runset = RunSet(args=args, chains=chains, chain_ids=chain_ids)
                 runset._csv_files = sample_csv_files
                 sample_fit = CmdStanMCMC(runset)
-                sample_drawset = sample_fit.draws_as_dataframe()
+                sample_drawset = sample_fit.draws_pd()
         except ValueError as exc:
             raise ValueError(
                 'Invalid mcmc_sample, error:\n\t{}\n\t'
@@ -870,6 +920,7 @@ class CmdStanModel:
                 data=_data,
                 seed=seed,
                 output_dir=gq_output_dir,
+                sig_figs=sig_figs,
                 method_args=generate_quantities_args,
             )
             runset = RunSet(args=args, chains=chains, chain_ids=chain_ids)
@@ -894,6 +945,7 @@ class CmdStanModel:
         seed: int = None,
         inits: float = None,
         output_dir: str = None,
+        sig_figs: int = None,
         save_diagnostics: bool = False,
         algorithm: str = None,
         iter: int = None,
@@ -948,6 +1000,11 @@ class CmdStanModel:
             files are written. If unspecified, output files will be written
             to a temporary directory which is deleted upon session exit.
 
+        :param sig_figs: Numerical precision used for output CSV and text files.
+            Must be an integer between 1 and 18.  If unspecified, the default
+            precision for the system file I/O is used; the usual value is 6.
+            Introduced in CmdStan-2.25.
+
         :param save_diagnostics: Whether or not to save diagnostics. If True,
             csv output files are written to an output file with filename
             template '<model_name>-<YYYYMMDDHHMM>-diagnostic-<chain_id>',
@@ -961,7 +1018,7 @@ class CmdStanModel:
 
         :param elbo_samples: Number of MC draws for estimate of ELBO.
 
-        :param eta: Stepsize scaling parameter.
+        :param eta: Step size scaling parameter.
 
         :param adapt_engaged: Whether eta adaptation is engaged.
 
@@ -1001,6 +1058,7 @@ class CmdStanModel:
                 seed=seed,
                 inits=_inits,
                 output_dir=output_dir,
+                sig_figs=sig_figs,
                 save_diagnostics=save_diagnostics,
                 method_args=variational_args,
             )
