@@ -1,19 +1,23 @@
 """CmdStan argument tests"""
 
+import logging
 import os
 import platform
 import unittest
 from time import time
 
-from cmdstanpy import _TMPDIR
+from testfixtures import LogCapture
+
+from cmdstanpy import _TMPDIR, cmdstan_path
 from cmdstanpy.cmdstan_args import (
-    Method,
-    SamplerArgs,
     CmdStanArgs,
-    OptimizeArgs,
     GenerateQuantitiesArgs,
+    Method,
+    OptimizeArgs,
+    SamplerArgs,
     VariationalArgs,
 )
+from cmdstanpy.utils import cmdstan_version_at
 
 HERE = os.path.dirname(os.path.abspath(__file__))
 DATAFILES_PATH = os.path.join(HERE, 'data')
@@ -571,6 +575,56 @@ class CmdStanArgsTest(unittest.TestCase):
                     method_args=sampler_args,
                 )
 
+    def test_args_sig_figs(self):
+        sampler_args = SamplerArgs()
+        cmdstan_path()  # sets os.environ['CMDSTAN']
+        if not cmdstan_version_at(2, 25):
+            with LogCapture() as log:
+                logging.getLogger()
+                CmdStanArgs(
+                    model_name='bernoulli',
+                    model_exe='bernoulli.exe',
+                    chain_ids=[1, 2, 3, 4],
+                    sig_figs=12,
+                    method_args=sampler_args,
+                )
+            expect = (
+                'arg sig_figs not valid, CmdStan version must be 2.25 '
+                'or higher, using verson {} in directory {}'
+            ).format(
+                os.path.basename(cmdstan_path()),
+                os.path.dirname(cmdstan_path()),
+            )
+            log.check_present(('cmdstanpy', 'WARNING', expect))
+        else:
+            cmdstan_args = CmdStanArgs(
+                model_name='bernoulli',
+                model_exe='bernoulli.exe',
+                chain_ids=[1, 2, 3, 4],
+                sig_figs=12,
+                method_args=sampler_args,
+            )
+            cmd = cmdstan_args.compose_command(
+                idx=0, csv_file='bern-output-1.csv'
+            )
+            self.assertIn('sig_figs=', ' '.join(cmd))
+            with self.assertRaises(ValueError):
+                CmdStanArgs(
+                    model_name='bernoulli',
+                    model_exe='bernoulli.exe',
+                    chain_ids=[1, 2, 3, 4],
+                    sig_figs=-1,
+                    method_args=sampler_args,
+                )
+            with self.assertRaises(ValueError):
+                CmdStanArgs(
+                    model_name='bernoulli',
+                    model_exe='bernoulli.exe',
+                    chain_ids=[1, 2, 3, 4],
+                    sig_figs=20,
+                    method_args=sampler_args,
+                )
+
 
 class GenerateQuantitesTest(unittest.TestCase):
     def test_args_fitted_params(self):
