@@ -4,9 +4,9 @@ import os
 import unittest
 
 from cmdstanpy import _TMPDIR
-from cmdstanpy.cmdstan_args import SamplerArgs, CmdStanArgs
-from cmdstanpy.utils import EXTENSION
+from cmdstanpy.cmdstan_args import CmdStanArgs, SamplerArgs
 from cmdstanpy.stanfit import RunSet
+from cmdstanpy.utils import EXTENSION
 
 HERE = os.path.dirname(os.path.abspath(__file__))
 DATAFILES_PATH = os.path.join(HERE, 'data')
