@@ -1,22 +1,28 @@
 """CmdStan method sample tests"""
 
+import logging
 import os
 import platform
-import logging
 import shutil
-from multiprocessing import cpu_count
-import tempfile
 import stat
+import tempfile
 import unittest
+from multiprocessing import cpu_count
 from time import time
-from testfixtures import LogCapture
+
 import pytest
+from testfixtures import LogCapture
+
+try:
+    import ujson as json
+except ImportError:
+    import json
 
 from cmdstanpy import _TMPDIR
-from cmdstanpy.cmdstan_args import Method, SamplerArgs, CmdStanArgs
-from cmdstanpy.utils import EXTENSION
-from cmdstanpy.stanfit import RunSet, CmdStanMCMC
+from cmdstanpy.cmdstan_args import CmdStanArgs, Method, SamplerArgs
 from cmdstanpy.model import CmdStanModel
+from cmdstanpy.stanfit import CmdStanMCMC, RunSet
+from cmdstanpy.utils import EXTENSION, cmdstan_version_at
 
 HERE = os.path.dirname(os.path.abspath(__file__))
 DATAFILES_PATH = os.path.join(HERE, 'data')
@@ -69,16 +75,21 @@ class SampleTest(unittest.TestCase):
             self.assertTrue(os.path.exists(csv_file))
             self.assertTrue(os.path.exists(stdout_file))
 
-        self.assertEqual(bern_fit.runset.chains, 2)
-        self.assertEqual(bern_fit.num_draws, 100)
+        self.assertEqual(bern_fit.chains, 2)
+        self.assertEqual(bern_fit.thin, 1)
+        self.assertEqual(bern_fit.num_draws_warmup, 1000)
+        self.assertEqual(bern_fit.num_draws_sampling, 100)
         self.assertEqual(bern_fit.column_names, tuple(BERNOULLI_COLS))
 
-        bern_draws = bern_fit.draws()
-        self.assertEqual(bern_draws.shape, (100, 2, len(BERNOULLI_COLS)))
+        self.assertEqual(bern_fit.draws().shape, (100, 2, len(BERNOULLI_COLS)))
         self.assertEqual(bern_fit.metric_type, 'diag_e')
-        self.assertEqual(bern_fit.stepsize.shape, (2,))
+        self.assertEqual(bern_fit.step_size.shape, (2,))
         self.assertEqual(bern_fit.metric.shape, (2, 1))
 
+        self.assertEqual(
+            bern_fit.draws(concat_chains=True).shape, (200, len(BERNOULLI_COLS))
+        )
+
         bern_fit = bern_model.sample(
             data=jdata,
             chains=2,
@@ -100,13 +111,13 @@ class SampleTest(unittest.TestCase):
             self.assertTrue(os.path.exists(stdout_file))
 
         self.assertEqual(bern_fit.runset.chains, 2)
-        self.assertEqual(bern_fit.num_draws, 100)
+        self.assertEqual(bern_fit.num_draws_sampling, 100)
         self.assertEqual(bern_fit.column_names, tuple(BERNOULLI_COLS))
 
         bern_sample = bern_fit.sample
         self.assertEqual(bern_sample.shape, (100, 2, len(BERNOULLI_COLS)))
         self.assertEqual(bern_fit.metric_type, 'dense_e')
-        self.assertEqual(bern_fit.stepsize.shape, (2,))
+        self.assertEqual(bern_fit.step_size.shape, (2,))
         self.assertEqual(bern_fit.metric.shape, (2, 1, 1))
 
         bern_fit = bern_model.sample(
@@ -122,8 +133,7 @@ class SampleTest(unittest.TestCase):
             stdout_file = bern_fit.runset.stdout_files[i]
             self.assertTrue(os.path.exists(csv_file))
             self.assertTrue(os.path.exists(stdout_file))
-        bern_draws = bern_fit.draws()
-        self.assertEqual(bern_draws.shape, (100, 2, len(BERNOULLI_COLS)))
+        self.assertEqual(bern_fit.draws().shape, (100, 2, len(BERNOULLI_COLS)))
         for i in range(bern_fit.runset.chains):  # cleanup datafile_path dir
             os.remove(bern_fit.runset.csv_files[i])
             if os.path.exists(bern_fit.runset.stdout_files[i]):
@@ -138,8 +148,7 @@ class SampleTest(unittest.TestCase):
             seed=12345,
             iter_sampling=100,
         )
-        bern_draws = bern_fit.draws()
-        self.assertEqual(bern_draws.shape, (100, 2, len(BERNOULLI_COLS)))
+        self.assertEqual(bern_fit.draws().shape, (100, 2, len(BERNOULLI_COLS)))
 
         data_dict = {'N': 10, 'y': [0, 1, 0, 0, 0, 0, 0, 0, 0, 1]}
         bern_fit = bern_model.sample(
@@ -149,8 +158,7 @@ class SampleTest(unittest.TestCase):
             seed=12345,
             iter_sampling=100,
         )
-        bern_draws = bern_fit.draws()
-        self.assertEqual(bern_draws.shape, (100, 2, len(BERNOULLI_COLS)))
+        self.assertEqual(bern_fit.draws().shape, (100, 2, len(BERNOULLI_COLS)))
 
     def test_init_types(self):
         stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
@@ -177,6 +185,40 @@ class SampleTest(unittest.TestCase):
         )
         self.assertIn('init=1', bern_fit.runset.__repr__())
 
+        # Save init to json
+        inits_path1 = os.path.join(_TMPDIR, 'inits_test_1.json')
+        with open(inits_path1, 'w') as fd:
+            json.dump({'theta': 0.1}, fd)
+        inits_path2 = os.path.join(_TMPDIR, 'inits_test_2.json')
+        with open(inits_path2, 'w') as fd:
+            json.dump({'theta': 0.9}, fd)
+
+        bern_fit = bern_model.sample(
+            data=jdata,
+            chains=2,
+            parallel_chains=2,
+            seed=12345,
+            iter_sampling=100,
+            inits=inits_path1,
+        )
+        self.assertIn(
+            'init={}'.format(inits_path1.replace('\\', '\\\\')),
+            bern_fit.runset.__repr__(),
+        )
+
+        bern_fit = bern_model.sample(
+            data=jdata,
+            chains=2,
+            parallel_chains=2,
+            seed=12345,
+            iter_sampling=100,
+            inits=[inits_path1, inits_path2],
+        )
+        self.assertIn(
+            'init={}'.format(inits_path1.replace('\\', '\\\\')),
+            bern_fit.runset.__repr__(),
+        )
+
         with self.assertRaises(ValueError):
             bern_model.sample(
                 data=jdata,
@@ -301,7 +343,7 @@ class SampleTest(unittest.TestCase):
             logistic_model.sample(
                 data=logistic_data, chains=7, threads_per_chain=5
             )
-            cores = cpu_count()
+            cores = max(min(cpu_count(), 7), 1)
             expect = 'total threads: {}'.format(cores * 5)
         log.check_present(('cmdstanpy', 'DEBUG', expect))
         with self.assertRaisesRegex(
@@ -327,7 +369,7 @@ class SampleTest(unittest.TestCase):
         self.assertEqual(datagen_fit.runset._args.method, Method.SAMPLE)
         self.assertEqual(datagen_fit.metric_type, None)
         self.assertEqual(datagen_fit.metric, None)
-        self.assertEqual(datagen_fit.stepsize, None)
+        self.assertEqual(datagen_fit.step_size, None)
 
         for i in range(datagen_fit.runset.chains):
             csv_file = datagen_fit.runset.csv_files[i]
@@ -341,95 +383,95 @@ class SampleTest(unittest.TestCase):
             'lp__',
             'accept_stat__',
             'N',
-            'y_sim.1',
-            'y_sim.2',
-            'y_sim.3',
-            'y_sim.4',
-            'y_sim.5',
-            'y_sim.6',
-            'y_sim.7',
-            'y_sim.8',
-            'y_sim.9',
-            'y_sim.10',
-            'y_sim.11',
-            'y_sim.12',
-            'y_sim.13',
-            'y_sim.14',
-            'y_sim.15',
-            'y_sim.16',
-            'y_sim.17',
-            'y_sim.18',
-            'y_sim.19',
-            'y_sim.20',
-            'x_sim.1',
-            'x_sim.2',
-            'x_sim.3',
-            'x_sim.4',
-            'x_sim.5',
-            'x_sim.6',
-            'x_sim.7',
-            'x_sim.8',
-            'x_sim.9',
-            'x_sim.10',
-            'x_sim.11',
-            'x_sim.12',
-            'x_sim.13',
-            'x_sim.14',
-            'x_sim.15',
-            'x_sim.16',
-            'x_sim.17',
-            'x_sim.18',
-            'x_sim.19',
-            'x_sim.20',
-            'pop_sim.1',
-            'pop_sim.2',
-            'pop_sim.3',
-            'pop_sim.4',
-            'pop_sim.5',
-            'pop_sim.6',
-            'pop_sim.7',
-            'pop_sim.8',
-            'pop_sim.9',
-            'pop_sim.10',
-            'pop_sim.11',
-            'pop_sim.12',
-            'pop_sim.13',
-            'pop_sim.14',
-            'pop_sim.15',
-            'pop_sim.16',
-            'pop_sim.17',
-            'pop_sim.18',
-            'pop_sim.19',
-            'pop_sim.20',
+            'y_sim[1]',
+            'y_sim[2]',
+            'y_sim[3]',
+            'y_sim[4]',
+            'y_sim[5]',
+            'y_sim[6]',
+            'y_sim[7]',
+            'y_sim[8]',
+            'y_sim[9]',
+            'y_sim[10]',
+            'y_sim[11]',
+            'y_sim[12]',
+            'y_sim[13]',
+            'y_sim[14]',
+            'y_sim[15]',
+            'y_sim[16]',
+            'y_sim[17]',
+            'y_sim[18]',
+            'y_sim[19]',
+            'y_sim[20]',
+            'x_sim[1]',
+            'x_sim[2]',
+            'x_sim[3]',
+            'x_sim[4]',
+            'x_sim[5]',
+            'x_sim[6]',
+            'x_sim[7]',
+            'x_sim[8]',
+            'x_sim[9]',
+            'x_sim[10]',
+            'x_sim[11]',
+            'x_sim[12]',
+            'x_sim[13]',
+            'x_sim[14]',
+            'x_sim[15]',
+            'x_sim[16]',
+            'x_sim[17]',
+            'x_sim[18]',
+            'x_sim[19]',
+            'x_sim[20]',
+            'pop_sim[1]',
+            'pop_sim[2]',
+            'pop_sim[3]',
+            'pop_sim[4]',
+            'pop_sim[5]',
+            'pop_sim[6]',
+            'pop_sim[7]',
+            'pop_sim[8]',
+            'pop_sim[9]',
+            'pop_sim[10]',
+            'pop_sim[11]',
+            'pop_sim[12]',
+            'pop_sim[13]',
+            'pop_sim[14]',
+            'pop_sim[15]',
+            'pop_sim[16]',
+            'pop_sim[17]',
+            'pop_sim[18]',
+            'pop_sim[19]',
+            'pop_sim[20]',
             'alpha_sim',
             'beta_sim',
-            'eta.1',
-            'eta.2',
-            'eta.3',
-            'eta.4',
-            'eta.5',
-            'eta.6',
-            'eta.7',
-            'eta.8',
-            'eta.9',
-            'eta.10',
-            'eta.11',
-            'eta.12',
-            'eta.13',
-            'eta.14',
-            'eta.15',
-            'eta.16',
-            'eta.17',
-            'eta.18',
-            'eta.19',
-            'eta.20',
+            'eta[1]',
+            'eta[2]',
+            'eta[3]',
+            'eta[4]',
+            'eta[5]',
+            'eta[6]',
+            'eta[7]',
+            'eta[8]',
+            'eta[9]',
+            'eta[10]',
+            'eta[11]',
+            'eta[12]',
+            'eta[13]',
+            'eta[14]',
+            'eta[15]',
+            'eta[16]',
+            'eta[17]',
+            'eta[18]',
+            'eta[19]',
+            'eta[20]',
         ]
         self.assertEqual(datagen_fit.column_names, tuple(column_names))
-        self.assertEqual(datagen_fit.num_draws, 100)
+        self.assertEqual(datagen_fit.num_draws_sampling, 100)
         self.assertEqual(datagen_fit.draws().shape, (100, 1, len(column_names)))
         self.assertEqual(datagen_fit.metric, None)
         self.assertEqual(datagen_fit.metric_type, None)
-        self.assertEqual(datagen_fit.stepsize, None)
+        self.assertEqual(datagen_fit.step_size, None)
 
     def test_bernoulli_file_with_space(self):
         self.test_bernoulli_good('bernoulli with space in name.stan')
@@ -441,6 +483,7 @@ class SampleTest(unittest.TestCase):
 
 
 class CmdStanMCMCTest(unittest.TestCase):
+    # pylint: disable=too-many-public-methods
     def test_validate_good_run(self):
         # construct fit using existing sampler output
         exe = os.path.join(DATAFILES_PATH, 'bernoulli' + EXTENSION)
@@ -471,14 +514,15 @@ class CmdStanMCMCTest(unittest.TestCase):
         self.assertTrue(runset._check_retcodes())
 
         fit = CmdStanMCMC(runset)
-        self.assertEqual(100, fit.num_draws)
+        self.assertEqual(1000, fit.num_draws_warmup)
+        self.assertEqual(100, fit.num_draws_sampling)
         self.assertEqual(len(BERNOULLI_COLS), len(fit.column_names))
         self.assertEqual('lp__', fit.column_names[0])
 
-        draws_df = fit.draws_as_dataframe()
+        draws_pd = fit.draws_pd()
         self.assertEqual(
-            draws_df.shape,
-            (fit.runset.chains * fit.num_draws, len(fit.column_names)),
+            draws_pd.shape,
+            (fit.runset.chains * fit.num_draws_sampling, len(fit.column_names)),
         )
 
         summary = fit.summary()
@@ -527,26 +571,18 @@ class CmdStanMCMCTest(unittest.TestCase):
             os.path.join(DATAFILES_PATH, 'runset-big', 'output_icar_nyc-1.csv'),
         ]
         fit = CmdStanMCMC(runset)
-        phis = ['phi.{}'.format(str(x + 1)) for x in range(2095)]
+        phis = ['phi[{}]'.format(str(x + 1)) for x in range(2095)]
         column_names = SAMPLER_STATE + phis
-        self.assertEqual(fit.num_draws, 1000)
+        self.assertEqual(fit.num_draws_sampling, 1000)
         self.assertEqual(fit.column_names, tuple(column_names))
         self.assertEqual(fit.metric_type, 'diag_e')
-        self.assertEqual(fit.stepsize.shape, (2,))
+        self.assertEqual(fit.step_size.shape, (2,))
         self.assertEqual(fit.metric.shape, (2, 2095))
         self.assertEqual((1000, 2, 2102), fit.draws().shape)
-        phis = fit.draws_as_dataframe(params=['phi'])
+        phis = fit.draws_pd(params=['phi'])
         self.assertEqual((2000, 2095), phis.shape)
-        phi1 = fit.draws_as_dataframe(params=['phi.1'])
-        self.assertEqual((2000, 1), phi1.shape)
-        mo_phis = fit.draws_as_dataframe(params=['phi.1', 'phi.10', 'phi.100'])
-        self.assertEqual((2000, 3), mo_phis.shape)
-        phi2095 = fit.draws_as_dataframe(params=['phi.2095'])
-        self.assertEqual((2000, 1), phi2095.shape)
-        with self.assertRaisesRegex(ValueError, 'unknown parameter: phi.2096'):
-            fit.draws_as_dataframe(params=['phi.2096'])
-        with self.assertRaisesRegex(ValueError, 'unknown parameter: ph'):
-            fit.draws_as_dataframe(params=['ph'])
+        with self.assertRaisesRegex(ValueError, r'unknown parameter: gamma'):
+            fit.draws_pd(params=['gamma'])
 
     # pylint: disable=no-self-use
     def test_custom_metric(self):
@@ -564,6 +600,52 @@ class CmdStanMCMCTest(unittest.TestCase):
             metric=jmetric,
         )
 
+        jmetric2 = os.path.join(DATAFILES_PATH, 'bernoulli.metric-2.json')
+        bern_model.sample(
+            data=jdata,
+            chains=2,
+            parallel_chains=2,
+            seed=12345,
+            iter_sampling=200,
+            metric=[jmetric, jmetric2],
+        )
+
+    def test_custom_step_size(self):
+        stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
+        jdata = os.path.join(DATAFILES_PATH, 'bernoulli.data.json')
+        bern_model = CmdStanModel(stan_file=stan)
+        # just test that it runs without error
+        bern_model.sample(
+            data=jdata,
+            chains=2,
+            parallel_chains=2,
+            seed=12345,
+            iter_sampling=200,
+            step_size=1,
+        )
+
+        bern_model.sample(
+            data=jdata,
+            chains=2,
+            parallel_chains=2,
+            seed=12345,
+            iter_sampling=200,
+            step_size=[1, 2],
+        )
+
+    def test_custom_seed(self):
+        stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
+        jdata = os.path.join(DATAFILES_PATH, 'bernoulli.data.json')
+        bern_model = CmdStanModel(stan_file=stan)
+        # just test that it runs without error
+        bern_model.sample(
+            data=jdata,
+            chains=2,
+            parallel_chains=2,
+            seed=[44444, 55555],
+            iter_sampling=200,
+        )
+
     def test_adapt_schedule(self):
         stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
         jdata = os.path.join(DATAFILES_PATH, 'bernoulli.data.json')
@@ -768,11 +850,33 @@ class CmdStanMCMCTest(unittest.TestCase):
             save_warmup=True,
         )
         self.assertEqual(bern_fit.column_names, tuple(BERNOULLI_COLS))
-        self.assertEqual(bern_fit.num_draws, 300)
+        self.assertEqual(bern_fit.num_draws_warmup, 200)
+        self.assertEqual(bern_fit.num_draws_sampling, 100)
         self.assertEqual(bern_fit.draws().shape, (100, 2, len(BERNOULLI_COLS)))
+        self.assertEqual(
+            bern_fit.draws(inc_warmup=False).shape,
+            (100, 2, len(BERNOULLI_COLS)),
+        )
+        self.assertEqual(
+            bern_fit.draws(concat_chains=True).shape,
+            (200, len(BERNOULLI_COLS)),
+        )
         self.assertEqual(
             bern_fit.draws(inc_warmup=True).shape, (300, 2, len(BERNOULLI_COLS))
         )
+        self.assertEqual(
+            bern_fit.draws(inc_warmup=True, concat_chains=True).shape,
+            (600, len(BERNOULLI_COLS)),
+        )
+
+        self.assertEqual(bern_fit.draws_pd().shape, (200, len(BERNOULLI_COLS)))
+        self.assertEqual(
+            bern_fit.draws_pd(inc_warmup=False).shape,
+            (200, len(BERNOULLI_COLS)),
+        )
+        self.assertEqual(
+            bern_fit.draws_pd(inc_warmup=True).shape, (600, len(BERNOULLI_COLS))
+        )
 
     def test_save_warmup_thin(self):
         stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
@@ -789,8 +893,10 @@ class CmdStanMCMCTest(unittest.TestCase):
             save_warmup=True,
         )
         self.assertEqual(bern_fit.column_names, tuple(BERNOULLI_COLS))
-        self.assertEqual(bern_fit.num_draws, 60)
         self.assertEqual(bern_fit.draws().shape, (20, 2, len(BERNOULLI_COLS)))
+        self.assertEqual(
+            bern_fit.draws(concat_chains=True).shape, (40, len(BERNOULLI_COLS))
+        )
         self.assertEqual(
             bern_fit.draws(inc_warmup=True).shape, (60, 2, len(BERNOULLI_COLS))
         )
@@ -809,7 +915,7 @@ class CmdStanMCMCTest(unittest.TestCase):
             save_warmup=False,
         )
         self.assertEqual(bern_fit.column_names, tuple(BERNOULLI_COLS))
-        self.assertEqual(bern_fit.num_draws, 100)
+        self.assertEqual(bern_fit.num_draws_sampling, 100)
         self.assertEqual(bern_fit.draws().shape, (100, 2, len(BERNOULLI_COLS)))
         with LogCapture() as log:
             self.assertEqual(
@@ -824,6 +930,32 @@ class CmdStanMCMCTest(unittest.TestCase):
                 ' must run sampler with "save_warmup=True".',
             )
         )
+        with LogCapture() as log:
+            self.assertEqual(
+                bern_fit.draws(inc_warmup=True, concat_chains=True).shape,
+                (200, len(BERNOULLI_COLS)),
+            )
+        log.check_present(
+            (
+                'cmdstanpy',
+                'WARNING',
+                'draws from warmup iterations not available,'
+                ' must run sampler with "save_warmup=True".',
+            )
+        )
+        with LogCapture() as log:
+            self.assertEqual(
+                bern_fit.draws_pd(inc_warmup=True).shape,
+                (200, len(BERNOULLI_COLS)),
+            )
+        log.check_present(
+            (
+                'cmdstanpy',
+                'WARNING',
+                'draws from warmup iterations not available,'
+                ' must run sampler with "save_warmup=True".',
+            )
+        )
 
     def test_deprecated(self):
         stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
@@ -840,8 +972,7 @@ class CmdStanMCMCTest(unittest.TestCase):
         )
         with LogCapture() as log:
             self.assertEqual(
-                bern_fit.sample.shape,
-                (100, 2, len(BERNOULLI_COLS)),
+                bern_fit.sample.shape, (100, 2, len(BERNOULLI_COLS))
             )
         log.check_present(
             (
@@ -853,8 +984,7 @@ class CmdStanMCMCTest(unittest.TestCase):
         )
         with LogCapture() as log:
             self.assertEqual(
-                bern_fit.warmup.shape,
-                (300, 2, len(BERNOULLI_COLS)),
+                bern_fit.warmup.shape, (300, 2, len(BERNOULLI_COLS))
             )
         log.check_present(
             (
@@ -873,10 +1003,30 @@ class CmdStanMCMCTest(unittest.TestCase):
         bern_fit = bern_model.sample(
             data=jdata, chains=2, seed=12345, iter_warmup=100, iter_sampling=100
         )
-        diags = bern_fit.sampler_diagnostics()
+        diags = bern_fit.sampler_variables()
         self.assertEqual(SAMPLER_STATE, list(diags))
         for key in diags:
             self.assertEqual(diags[key].shape, (100, 2))
+            self.assertEqual(
+                bern_fit.sample.shape, (100, 2, len(BERNOULLI_COLS))
+            )
+
+        with LogCapture() as log:
+            diags = bern_fit.sampler_diagnostics()
+            self.assertEqual(SAMPLER_STATE, list(diags))
+            for key in diags:
+                self.assertEqual(diags[key].shape, (100, 2))
+                self.assertEqual(
+                    bern_fit.sample.shape, (100, 2, len(BERNOULLI_COLS))
+                    )
+        log.check_present(
+            (
+                'cmdstanpy',
+                'WARNING',
+                'method "sample" will be deprecated,'
+                ' use method "draws" instead.',
+            )
+        )
 
     def test_variable_bern(self):
         stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
@@ -885,9 +1035,9 @@ class CmdStanMCMCTest(unittest.TestCase):
         bern_fit = bern_model.sample(
             data=jdata, chains=2, seed=12345, iter_warmup=100, iter_sampling=100
         )
-        self.assertEqual(1, len(bern_fit._stan_variable_dims))
-        self.assertTrue('theta' in bern_fit._stan_variable_dims)
-        self.assertEqual(bern_fit._stan_variable_dims['theta'], 1)
+        self.assertEqual(1, len(bern_fit.stan_vars_dims))
+        self.assertTrue('theta' in bern_fit.stan_vars_dims)
+        self.assertEqual(bern_fit.stan_vars_dims['theta'], ())
         theta = bern_fit.stan_variable(name='theta')
         self.assertEqual(theta.shape, (200,))
         with self.assertRaises(ValueError):
@@ -895,7 +1045,7 @@ class CmdStanMCMCTest(unittest.TestCase):
         with self.assertRaises(ValueError):
             bern_fit.stan_variable(name='lp__')
 
-    def test_variable_lv(self):
+    def test_variables_2d(self):
         # pylint: disable=C0103
         # construct fit using existing sampler output
         exe = os.path.join(DATAFILES_PATH, 'lotka-volterra' + EXTENSION)
@@ -914,22 +1064,24 @@ class CmdStanMCMCTest(unittest.TestCase):
         runset._csv_files = [os.path.join(DATAFILES_PATH, 'lotka-volterra.csv')]
         runset._set_retcode(0, 0)
         fit = CmdStanMCMC(runset)
-        self.assertEqual(20, fit.num_draws)
-        self.assertEqual(8, len(fit._stan_variable_dims))
-        self.assertTrue('z' in fit._stan_variable_dims)
-        self.assertEqual(fit._stan_variable_dims['z'], (20, 2))
-        z = fit.stan_variable(name='z')
-        self.assertEqual(z.shape, (20, 20, 2))
-        theta = fit.stan_variable(name='theta')
-        self.assertEqual(theta.shape, (20, 4))
+        self.assertEqual(20, fit.num_draws_sampling)
+        self.assertEqual(8, len(fit.stan_vars_dims))
+        self.assertTrue('z' in fit.stan_vars_dims)
+        self.assertEqual(fit.stan_vars_dims['z'], (20, 2))
+        vars = fit.stan_variables()
+        self.assertEqual(len(vars), len(fit.stan_vars_dims))
+        self.assertTrue('z' in vars)
+        self.assertEqual(vars['z'].shape, (20, 20, 2))
+        self.assertTrue('theta' in vars)
+        self.assertEqual(vars['theta'].shape, (20, 4))
 
-    def test_variables(self):
+    def test_variables_3d(self):
         # construct fit using existing sampler output
-        exe = os.path.join(DATAFILES_PATH, 'lotka-volterra' + EXTENSION)
-        jdata = os.path.join(DATAFILES_PATH, 'lotka-volterra.data.json')
+        exe = os.path.join(DATAFILES_PATH, 'multidim_vars' + EXTENSION)
+        jdata = os.path.join(DATAFILES_PATH, 'logistic.data.R')
         sampler_args = SamplerArgs(iter_sampling=20)
         cmdstan_args = CmdStanArgs(
-            model_name='lotka-volterra',
+            model_name='multidim_vars',
             model_exe=exe,
             chain_ids=[1],
             seed=12345,
@@ -938,19 +1090,27 @@ class CmdStanMCMCTest(unittest.TestCase):
             method_args=sampler_args,
         )
         runset = RunSet(args=cmdstan_args, chains=1)
-        runset._csv_files = [os.path.join(DATAFILES_PATH, 'lotka-volterra.csv')]
+        runset._csv_files = [os.path.join(DATAFILES_PATH, 'multidim_vars.csv')]
         runset._set_retcode(0, 0)
         fit = CmdStanMCMC(runset)
-        self.assertEqual(20, fit.num_draws)
-        self.assertEqual(8, len(fit._stan_variable_dims))
-        self.assertTrue('z' in fit._stan_variable_dims)
-        self.assertEqual(fit._stan_variable_dims['z'], (20, 2))
+        self.assertEqual(20, fit.num_draws_sampling)
+        self.assertEqual(3, len(fit.stan_vars_dims))
+        self.assertTrue('y_rep' in fit.stan_vars_dims)
+        self.assertEqual(fit.stan_vars_dims['y_rep'], (5, 4, 3))
+        var_y_rep = fit.stan_variable(name='y_rep')
+        self.assertEqual(var_y_rep.shape, (20, 5, 4, 3))
+        var_beta = fit.stan_variable(name='beta')
+        self.assertEqual(var_beta.shape, (20, 2))
+        var_frac_60 = fit.stan_variable(name='frac_60')
+        self.assertEqual(var_frac_60.shape, (20,))
         vars = fit.stan_variables()
-        self.assertEqual(len(vars), len(fit._stan_variable_dims))
-        self.assertTrue('z' in vars)
-        self.assertEqual(vars['z'].shape, (20, 20, 2))
-        self.assertTrue('theta' in vars)
-        self.assertEqual(vars['theta'].shape, (20, 4))
+        self.assertEqual(len(vars), len(fit.stan_vars_dims))
+        self.assertTrue('y_rep' in vars)
+        self.assertEqual(vars['y_rep'].shape, (20, 5, 4, 3))
+        self.assertTrue('beta' in vars)
+        self.assertEqual(vars['beta'].shape, (20, 2))
+        self.assertTrue('frac_60' in vars)
+        self.assertEqual(vars['frac_60'].shape, (20,))
 
     def test_validate(self):
         stan = os.path.join(DATAFILES_PATH, 'bernoulli.stan')
@@ -976,7 +1136,21 @@ class CmdStanMCMCTest(unittest.TestCase):
 
         with LogCapture() as log:
             logging.getLogger()
-            self.assertIsNone(bern_fit.stan_variable_dims)
+            self.assertIsNone(bern_fit.num_unconstrained_params)
+        expect = 'csv files not yet validated'
+        msg = log.actual()[-1][-1]
+        self.assertTrue(msg.startswith(expect))
+
+        with LogCapture() as log:
+            logging.getLogger()
+            self.assertIsNone(bern_fit.stan_vars_dims)
+        expect = 'csv files not yet validated'
+        msg = log.actual()[-1][-1]
+        self.assertTrue(msg.startswith(expect))
+
+        with LogCapture() as log:
+            logging.getLogger()
+            self.assertIsNone(bern_fit.stan_vars_cols)
         expect = 'csv files not yet validated'
         msg = log.actual()[-1][-1]
         self.assertTrue(msg.startswith(expect))
@@ -997,19 +1171,173 @@ class CmdStanMCMCTest(unittest.TestCase):
 
         with LogCapture() as log:
             logging.getLogger()
-            self.assertIsNone(bern_fit.stepsize)
+            self.assertIsNone(bern_fit.step_size)
         expect = 'csv files not yet validated'
         msg = log.actual()[-1][-1]
         self.assertTrue(msg.startswith(expect))
 
         # check computations match
-        self.assertEqual(bern_fit.num_draws, 150)
+        self.assertEqual(bern_fit.num_draws_warmup, 100)
+        self.assertEqual(bern_fit.num_draws_sampling, 50)
         bern_fit.validate_csv_files()
-        self.assertEqual(bern_fit.num_draws, 150)
+        self.assertEqual(bern_fit.num_draws_warmup, 100)
+        self.assertEqual(bern_fit.num_draws_sampling, 50)
         self.assertEqual(len(bern_fit.column_names), 8)
-        self.assertEqual(len(bern_fit.stan_variable_dims), 1)
+        self.assertEqual(len(bern_fit.stan_vars_dims), 1)
+        self.assertEqual(len(bern_fit.stan_vars_cols.keys()), 1)
         self.assertEqual(bern_fit.metric_type, 'diag_e')
 
+    def test_validate_sample_sig_figs(self, stanfile='bernoulli.stan'):
+        if cmdstan_version_at(2, 25):
+            stan = os.path.join(DATAFILES_PATH, stanfile)
+            bern_model = CmdStanModel(stan_file=stan)
+
+            jdata = os.path.join(DATAFILES_PATH, 'bernoulli.data.json')
+            bern_fit = bern_model.sample(
+                data=jdata,
+                chains=1,
+                seed=12345,
+                iter_sampling=100,
+            )
+            bern_draws = bern_fit.draws()
+            theta = format(bern_draws[99, 0, 7], '.18g')
+            self.assertFalse(theta.startswith('0.21238045821757600'))
+
+            bern_fit_17 = bern_model.sample(
+                data=jdata,
+                chains=1,
+                seed=12345,
+                iter_sampling=100,
+                sig_figs=17,
+            )
+            self.assertTrue(bern_fit_17.draws().size)
+
+            with self.assertRaises(ValueError):
+                bern_model.sample(
+                    data=jdata,
+                    chains=1,
+                    seed=12345,
+                    iter_sampling=100,
+                    sig_figs=27,
+                )
+                with self.assertRaises(ValueError):
+                    bern_model.sample(
+                        data=jdata,
+                        chains=1,
+                        seed=12345,
+                        iter_sampling=100,
+                        sig_figs=-1,
+                    )
+
+    def test_validate_summary_sig_figs(self):
+        # construct CmdStanMCMC from logistic model output, config
+        exe = os.path.join(DATAFILES_PATH, 'logistic' + EXTENSION)
+        rdata = os.path.join(DATAFILES_PATH, 'logistic.data.R')
+        sampler_args = SamplerArgs(iter_sampling=100)
+        cmdstan_args = CmdStanArgs(
+            model_name='logistic',
+            model_exe=exe,
+            chain_ids=[1, 2, 3, 4],
+            seed=12345,
+            data=rdata,
+            output_dir=DATAFILES_PATH,
+            sig_figs=17,
+            method_args=sampler_args,
+        )
+        runset = RunSet(args=cmdstan_args)
+        runset._csv_files = [
+            os.path.join(DATAFILES_PATH, 'logistic_output_1.csv'),
+            os.path.join(DATAFILES_PATH, 'logistic_output_2.csv'),
+            os.path.join(DATAFILES_PATH, 'logistic_output_3.csv'),
+            os.path.join(DATAFILES_PATH, 'logistic_output_4.csv'),
+        ]
+        retcodes = runset._retcodes
+        for i in range(len(retcodes)):
+            runset._set_retcode(i, 0)
+        fit = CmdStanMCMC(runset)
+
+        sum_default = fit.summary()
+        beta1_default = format(sum_default.iloc[1, 0], '.18g')
+        self.assertTrue(beta1_default.startswith('1.3'))
+
+        if cmdstan_version_at(2, 25):
+            sum_17 = fit.summary(sig_figs=17)
+            beta1_17 = format(sum_17.iloc[1, 0], '.18g')
+            self.assertTrue(beta1_17.startswith('1.345767078273'))
+
+            sum_10 = fit.summary(sig_figs=10)
+            beta1_10 = format(sum_10.iloc[1, 0], '.18g')
+            self.assertTrue(beta1_10.startswith('1.34576707'))
+
+        with self.assertRaises(ValueError):
+            fit.summary(sig_figs=20)
+        with self.assertRaises(ValueError):
+            fit.summary(sig_figs=-1)
+
+    def test_metadata(self):
+        # construct CmdStanMCMC from logistic model output, config
+        exe = os.path.join(DATAFILES_PATH, 'logistic' + EXTENSION)
+        rdata = os.path.join(DATAFILES_PATH, 'logistic.data.R')
+        sampler_args = SamplerArgs(iter_sampling=100)
+        cmdstan_args = CmdStanArgs(
+            model_name='logistic',
+            model_exe=exe,
+            chain_ids=[1, 2, 3, 4],
+            seed=12345,
+            data=rdata,
+            output_dir=DATAFILES_PATH,
+            sig_figs=17,
+            method_args=sampler_args,
+        )
+        runset = RunSet(args=cmdstan_args)
+        runset._csv_files = [
+            os.path.join(DATAFILES_PATH, 'logistic_output_1.csv'),
+            os.path.join(DATAFILES_PATH, 'logistic_output_2.csv'),
+            os.path.join(DATAFILES_PATH, 'logistic_output_3.csv'),
+            os.path.join(DATAFILES_PATH, 'logistic_output_4.csv'),
+        ]
+        retcodes = runset._retcodes
+        for i in range(len(retcodes)):
+            runset._set_retcode(i, 0)
+        fit = CmdStanMCMC(runset)
+
+        col_names = tuple(
+            [
+                'lp__',
+                'accept_stat__',
+                'stepsize__',
+                'treedepth__',
+                'n_leapfrog__',
+                'divergent__',
+                'energy__',
+                'beta[1]',
+                'beta[2]',
+            ]
+        )
+
+        self.assertEqual(fit.chains, 4)
+        self.assertEqual(fit.chain_ids, [1, 2, 3, 4])
+        self.assertEqual(fit.num_draws_warmup, 1000)
+        self.assertEqual(fit.num_draws_sampling, 100)
+        self.assertEqual(fit.column_names, col_names)
+        self.assertEqual(fit.num_unconstrained_params, 2)
+        self.assertEqual(fit.metric_type, 'diag_e')
+
+        self.assertEqual(fit.sampler_config['num_samples'], 100)
+        self.assertEqual(fit.sampler_config['thin'], 1)
+        self.assertEqual(fit.sampler_config['algorithm'], 'hmc')
+        self.assertEqual(fit.sampler_config['metric'], 'diag_e')
+        self.assertAlmostEqual(fit.sampler_config['delta'], 0.80)
+
+        self.assertTrue('n_leapfrog__' in fit.sampler_vars_cols)
+        self.assertTrue('energy__' in fit.sampler_vars_cols)
+        self.assertTrue('beta' not in fit.sampler_vars_cols)
+        self.assertTrue('energy__' not in fit.stan_vars_dims)
+        self.assertTrue('beta' in fit.stan_vars_dims)
+        self.assertTrue('beta' in fit.stan_vars_cols)
+        self.assertEqual(fit.stan_vars_dims['beta'], tuple([2]))
+        self.assertEqual(fit.stan_vars_cols['beta'], tuple([7, 8]))
+
 
 if __name__ == '__main__':
     unittest.main()
